@@ -96,6 +96,9 @@ static const struct argp_option args_options[] =
   {
     { "service", 's', N_("CONFIG"), 0, N_("Service configuration to be used") },
     { "no-idn", 'i', NULL, 0, N_("disable IDN encoding") },
+#ifdef __CYGWIN__
+    { "windows", 'w', NULL, 0, N_("Cygwin:Windows account mapping only (with passwd and group databases)") },
+#endif
     { NULL, 0, NULL, 0, NULL },
   };
 
@@ -117,6 +120,10 @@ static struct argp argp =
 /* Additional getaddrinfo flags for IDN encoding.  */
 static int idn_flags = AI_IDN | AI_CANONIDN;
 
+#ifdef __CYGWIN__
+static int windows_only = 0;
+#endif
+
 /* Print the version information.  */
 static void
 print_version (FILE *stream, struct argp_state *state)
@@ -221,32 +228,6 @@ ethers_keys (int number, char *key[])
 }
 #endif /* !__CYGWIN__ */
 
-/* This is for group */
-static void
-print_group (struct group *grp)
-{
-  unsigned int i = 0;
-
-#ifdef __CYGWIN__
-  /* Disable printing gid -1, the fallback invalid group called "????????" on
-     Cygwin versions prior to the introduction of direct SAM/AD access. */
-  if (grp->gr_gid == (gid_t) -1)
-    return;
-#endif
-  printf ("%s:%s:%lu:", grp->gr_name ? grp->gr_name : "",
-	  grp->gr_passwd ? grp->gr_passwd : "",
-	  (unsigned long int) grp->gr_gid);
-
-  while (grp->gr_mem[i] != NULL)
-    {
-      fputs_unlocked (grp->gr_mem[i], stdout);
-      ++i;
-      if (grp->gr_mem[i] != NULL)
-	putchar_unlocked (',');
-    }
-  putchar_unlocked ('\n');
-}
-
 #ifdef __CYGWIN__
 static PSID
 get_sid_from_special_string (const char *key, PBOOL try_as_cygname)
@@ -295,8 +276,61 @@ get_sid_from_special_string (const char *key, PBOOL try_as_cygname)
     }
   return sid;
 }
+
+static void
+print_windows_mapping (const char *cygname, unsigned long id,
+		       const char *sid_string)
+{
+  PSID sid;
+  WCHAR name[NAME_MAX + 1];
+  DWORD nlen = NAME_MAX + 1;
+  /* DNLEN is moot.  Domain names can now be twice as long. */
+  WCHAR dom[32];
+  DWORD dlen = 32;
+  SID_NAME_USE use;
+
+  if (!cygname || !sid_string)
+    return;
+  if (!ConvertStringSidToSidA (sid_string, &sid))
+    return;
+  if (LookupAccountSidW (NULL, sid, name, &nlen, dom, &dlen, &use))
+    printf ("%s:%lu:%ls\\%ls:%s\n", cygname, id, dom, name, sid_string);
+  LocalFree (sid);
+}
 #endif
 
+/* This is for group */
+static void
+print_group (struct group *grp)
+{
+  unsigned int i = 0;
+
+#ifdef __CYGWIN__
+  /* Disable printing gid -1, the fallback invalid group called "????????" on
+     Cygwin versions prior to the introduction of direct SAM/AD access. */
+  if (grp->gr_gid == (gid_t) -1)
+    return;
+  if (windows_only)
+    {
+      print_windows_mapping (grp->gr_name, (unsigned long) grp->gr_gid,
+			     grp->gr_passwd);
+      return;
+    }
+#endif
+  printf ("%s:%s:%lu:", grp->gr_name ? grp->gr_name : "",
+	  grp->gr_passwd ? grp->gr_passwd : "",
+	  (unsigned long int) grp->gr_gid);
+
+  while (grp->gr_mem[i] != NULL)
+    {
+      fputs_unlocked (grp->gr_mem[i], stdout);
+      ++i;
+      if (grp->gr_mem[i] != NULL)
+	putchar_unlocked (',');
+    }
+  putchar_unlocked ('\n');
+}
+
 static int
 group_keys (int number, char *key[])
 {
@@ -740,6 +774,12 @@ print_passwd (struct passwd *pwd)
      Cygwin versions prior to the introduction of direct SAM/AD access. */
   if (pwd->pw_uid == (uid_t) -1)
     return;
+  if (windows_only)
+    {
+      char *c = strstr (pwd->pw_gecos, "S-1-");
+      print_windows_mapping (pwd->pw_name, (unsigned long) pwd->pw_uid, c);
+      return;
+    }
 #endif
   printf ("%s:%s:%lu:%lu:%s:%s:%s\n",
 	  pwd->pw_name ? pwd->pw_name : "",
@@ -1095,6 +1135,12 @@ parse_option (int key, char *arg, struct argp_state *state)
       idn_flags = 0;
       break;
 
+#ifdef __CYGWIN__
+    case 'w':
+      windows_only = 1;
+      break;
+#endif
+
     default:
       return ARGP_ERR_UNKNOWN;
     }
